#pragma once
#include <winsock2.h>
#include <ws2ipdef.h>
#include <ws2tcpip.h>

#include <string.h>

#include "../../except.hpp"
#include "../../system.hpp"
#include "../socket.hpp"

namespace fcf {
namespace net {
namespace details {

  struct address_t {
  union {
      sockaddr     address;
      sockaddr_in  address4;
      sockaddr_in6 address6;
  };
  int size;

  address_t() {
      memset(this, 0, sizeof(address_t));
  }

  address_t(const char* a_addrstr, int a_port) {
      memset(this, 0, sizeof(address_t));

      bool isv4 = strchr(a_addrstr, ':') == 0;

      if (isv4) {
          if (a_addrstr[0] == 0)
              a_addrstr = "0.0.0.0";
          address4.sin_family = AF_INET;
          address4.sin_port = htons(a_port);
          if (inet_pton(AF_INET, a_addrstr, &address4.sin_addr) <= 0)
              throw fcf::except::incorrectipformat(a_addrstr);
          size = sizeof(sockaddr_in);
      }
      else {
          address6.sin6_family = AF_INET6;
          address6.sin6_port = htons(a_port);
          if (inet_pton(AF_INET6, a_addrstr, address6.sin6_addr.s6_addr) <= 0)
              throw fcf::except::incorrectipformat(a_addrstr);
          size = sizeof(sockaddr_in6);
      }
  }
  };


  class socket_t : public fcf::net::socket_t {
    public:
    socket_t(const char* a_address, int a_port)
      : _handle(0), _addr(a_address), _port(a_port) {
        address_t address(a_address, a_port);
      _family = address.address.sa_family;
      _handle = ::socket(address.address.sa_family == AF_INET ? PF_INET : PF_INET6, SOCK_STREAM, 0);
      if (_handle == -1 || !_handle) {
        _handle = 0;
        int err = WSAGetLastError();
        if (err == ENFILE || err == EMFILE)
          throw fcf::except::filelimit();
        else if (err == ENOBUFS || err == ENOMEM)
          throw fcf::except::nomem();
        else
          throw fcf::except::createsocket(fcf::get_system_error_str(err).c_str());
      }
    }

    socket_t(socket_handle_t a_socket_handle, const char* a_address, int a_port)
       : _handle(a_socket_handle), _addr(a_address), _port(a_port) {
    }

    virtual ~socket_t() {
      close();
    }

    virtual fcf::net::socket_t* accept() {
      address_t address;
      address.size = sizeof(address.address6);
      socket_handle_t handle = ::accept(_handle, &address.address, &address.size);
      if (handle == -1 || !handle) {
        int err = WSAGetLastError();
        if (err == EAGAIN || err == EWOULDBLOCK)
          throw fcf::except::socketagain();
        else if (err == EMFILE || err == ENFILE)
          throw fcf::except::filelimit();
        else
          throw std::runtime_error(fcf::get_system_error_str(err).c_str());
      }
      char addrbuf[INET6_ADDRSTRLEN + 1];
      memset(&addrbuf[0], 0, sizeof(addrbuf));
      void* addrptr;
      int   port;
      if (_family == AF_INET){
        addrptr = &address.address4.sin_addr;
        port    = ntohs(address.address4.sin_port);
      } else {
        addrptr = &address.address6.sin6_addr.s6_addr;
        port    = ntohs(address.address6.sin6_port);
      }
      inet_ntop(_family, addrptr, &addrbuf[0], sizeof(addrbuf));
      socket_t* ps = new socket_t(handle, &addrbuf[0], port);
      return ps;
    }

    virtual void connect() {
      address_t address(_addr.c_str(), _port);
      if (::connect(_handle, &address.address, address.size) && WSAGetLastError() != EINPROGRESS)
        throw fcf::except::connection_error(_addr.c_str(), _port, fcf::get_system_error_str(WSAGetLastError()).c_str());
    }

    virtual void bind() {
      address_t address(_addr.c_str(), _port);
      if (::bind(_handle, &address.address, address.size) != 0) {
        int err = WSAGetLastError();
        if (err == EACCES)
          throw fcf::except::accessbind(_addr.c_str(), _port);
        else if (err == EADDRINUSE)
          throw fcf::except::addrinuse(_addr.c_str(), _port);
        else if (err == EADDRINUSE)
          throw fcf::except::nomem();
        else
          throw fcf::except::errorbind(_addr.c_str(), _port, fcf::get_system_error_str(err).c_str());
      }
    }

    virtual void listen(int a_queueSize) {
      if (::listen(_handle, a_queueSize) != 0) {
        int err = WSAGetLastError();
        if (err == EADDRINUSE)
          throw fcf::except::addrinuse(_addr.c_str(), _port);
        else
          throw fcf::except::listenerror(_addr.c_str(), _port, fcf::get_system_error_str(err).c_str());
      }
    }

    virtual unsigned int send(const char* a_data, unsigned int a_size) {
      int ret = ::send(_handle, a_data, a_size, 0);
      if (ret < 0) {
        int err = WSAGetLastError();
        if (err == EAGAIN || err == EWOULDBLOCK)
          return 0;
        else if (err == ECONNRESET)
          throw fcf::except::connreset();
        else if (err == ECONNREFUSED)
          throw fcf::except::connrefused();
        else
          throw fcf::except::socketsend(fcf::get_system_error_str(err).c_str());
      } else if (ret == 0){
        throw fcf::except::socketempty();
      }

      return ret;
    }


    virtual unsigned int recv(char* a_buffer, unsigned int a_size) {
      int ret = ::recv(_handle, a_buffer, a_size, 0);
      if (ret < 0) {
        int err = WSAGetLastError();
        if (err == EAGAIN || err == EWOULDBLOCK)
          return 0;
        else if (err == ECONNRESET)
          throw fcf::except::connreset();
        else if (err == ECONNREFUSED)
          throw fcf::except::connrefused();
        else
          throw fcf::except::socketrec(fcf::get_system_error_str(err).c_str());
      } else if (ret == 0){
        throw fcf::except::socketempty();
      }

      return ret;
    }

    virtual void close() {
      if (!_handle)
        return;
      ::shutdown(_handle, SD_BOTH);
      ::closesocket(_handle);
      _handle = 0;
    }

    virtual void set_blocking(bool a_val) {
      unsigned long mode = a_val ? 0 : 1;
      ioctlsocket(_handle, FIONBIO, &mode);
    }

    virtual void set_timeout(int a_sec) {
      a_sec *= 1000;
      setsockopt(_handle, SOL_SOCKET, SO_RCVTIMEO, (const char*)&a_sec, sizeof(a_sec));
      setsockopt(_handle, SOL_SOCKET, SO_SNDTIMEO, (const char*)&a_sec, sizeof(a_sec));
    }

    virtual const char* get_address() {
      return _addr.c_str();
    }

    virtual int get_port() {
      return _port;
    }

    virtual int get_error() {
      int error = 0;
      socklen_t error_len = sizeof(error);
      getsockopt(_handle, SOCK_STREAM, SO_ERROR, (char*)&error, &error_len);
      return error;
    }

    virtual void set_reuse(bool a_val) {
      int mode = a_val ? 1 : 0;
      setsockopt(_handle, SOL_SOCKET, SO_REUSEADDR, (const char*)&mode, sizeof(mode));
    }

    virtual socket_handle_t get_handle() {
      return _handle;
    }

    private:
    socket_handle_t _handle;
    std::string     _addr;
    int             _port;
    int             _family;
  };

} // details namespace
} // net namespace
} // fcf namespace
